define('StreamXMLParser', ['exports'], function (exports) { 'use strict';

function ObjectAssign(target) {
	var arguments$1 = arguments;

	var hop = Object.prototype.hasOwnProperty;
	var argc = arguments.length;
	for (var i = 1; i < argc; i++) {
		var source = arguments$1[i];
		for (var key in source) {
			if (hop.call(source, key)) {
				target[key] = source[key];
			}
		}
	}
	return target;
}

var st_TEXT = {name:'TEXT'};
var st_TAG_START = {name:'TAG_START'};
var st_TAG_NAME = {name:'TAG_NAME'};
var st_TAG_NAMED = {name:'TAG_NAMED'};
var st_TAG_NAMED_SELFCLOSE = {name:'TAG_NAMED_SELFCLOSE'};
var st_ATTR_NAME = {name:'ATTR_NAME'};
var st_ATTR_NAMED = {name:'ATTR_NAMED'};
var st_ATTR_VALUE_START = {name:'ATTR_VALUE_START'};
var st_ATTR_VALUE_QUOTED = {name:'ATTR_VALUE_QUOTED'};
var st_ATTR_VALUE_RAW = {name:'ATTR_VALUE_RAW'};
var st_ATTR_VALUE_CUSTOM = {name:'ATTR_VALUE_CUSTOM'};
var st_INSTRUCTION = {name:'INSTRUCTION'};
var st_DECLARATION_START = {name:'DECLARATION_START'};
var st_DECLARATION = {name:'DECLARATION'};
var st_COMMENT = {name:'COMMENT'};
var st_CDATA = {name:'CDATA'};

var ev_text = {name:'text'};
var ev_startTag = {name:'startTag'};
var ev_endTag = {name:'endTag'};
var ev_startInstruction = {name:'startInstruction'};
var ev_endInstruction = {name:'endInstruction'};
var ev_startDeclaration = {name:'startDeclaration'};
var ev_endDeclaration = {name:'endDeclaration'};
var ev_startComment = {name:'startComment'};
var ev_endComment = {name:'endComment'};
var ev_startCdata = {name:'startCdata'};
var ev_endCdata = {name:'endCdata'};
var ev_tagName = {name:'tagName'};
var ev_tagAttribute = {name:'tagAttribute'};
var ev_info = {name:'info'};

var reSpace = /\s/;
var strCdata = '[CDATA[';
// var reCdata = /^\[CDATA\[$/i;
// var lenCdata = 7;
var echo$1 = function (x) { return x; };
var nop = function () {};

function XMLParser(opt) {
	if (opt instanceof Function) {
		this.event = opt;
		this.opt = opt = { event: opt };
	} else {
		this.event = opt.event;
		this.opt = opt;
	}
	var ref = ObjectAssign({}, XMLParser.optDefault,
		opt);
	var tagStrictMap = ref.tagStrictMap;
	var isStrictTag = ref.isStrictTag;
	if (tagStrictMap) { this.tagStrictMap = tagStrictMap; }
	if (isStrictTag) { this.isStrictTag = isStrictTag; }
	this.decodeString = opt.decodeString || echo$1;
	this.decodeText = opt.decodeText || this.decodeString;
	this.decodeTagName = opt.decodeTagName || this.decodeString;
	this.decodeAttrName = opt.decodeAttrName || this.decodeString;
	this.decodeAttrValue = opt.decodeAttrValue || this.decodeString;
	var cp = opt.customParser;
	this.getCustomParser = ObjectAssign({}, {attrValue: nop},
		cp);
	this.currentCustomParser = {
		attrValue: null
	};
	this.state = st_TEXT;
	this.buffer = '';
}

XMLParser.optDefault = {
	tagStrictMap: null,
	isStrictTag: null,
};

XMLParser.prototype = {
	constructor: XMLParser,
	opt: null,
	state: null,
	c: null,
	pos: 0,
	line: 0,
	column: 0,
	endPos: 0,
	endLine: 0,
	endColumn: 0,
	tagPath: null,
	currentTag: null,
	currentAttr: null,
	decodeString: null,
	decodeText: null,
	decodeTagName: null,
	decodeAttrName: null,
	decodeAttrValue: null,
	customParser: null,
	currentCustomParser: null,
	tagStrictMap: null,
	tagStrictOpen: null,
	tagStrictBuf: null,
	tagStrictEventBuf: null,
	buffer: null,
	isStrictTag: function(name) {
		var smap = this.tagStrictMap;
		return smap && smap[String(name).toLowerCase()];
	},
	createAttr: function() {
		return {
			startSpace: null,
			name: null,
			eqSpace: null,
			value: null,
			valueSpace: null,
			quotes: null
		};
	},
	decodeTag: function(ctag, ev) {
		var name = ctag.name;
		if ('string' === typeof name) {
			name = this.decodeTagName(name, ev);
		}
		return ObjectAssign({}, ctag, {name: name});
	},
	decodeAttr: function(cattr, ev) {
		var name = cattr.name;
		var value = cattr.value;
		if ('string' === typeof name) {
			name = this.decodeAttrName(name, ev);
		}
		if ('string' === typeof value) {
			value = this.decodeAttrValue(value, ev);
		}
		return ObjectAssign({}, cattr, {name: name, value: value});
	},
	write: function(text, final) {
		var assign;

		function getEndChars(n) {
			var j = i + 1;
			if (j >= n) {
				return text.substr(j - n, n);
			} else {
				var blen = buf.length;
				if (j + blen >= n) {
					return buf.substr(blen - (n - j), n - j).concat(text.substr(0, j))
				} else {
					throw new Error('Requested '+n+' end characters but there are '+j+' in text beginning and '+blen+' in buffer');
				}
			}
		}
		function event(srcEv, id, evInfo) {
			var ev = {
				name: srcEv.name,
				id: id,
				attr: cattr,
				tag: ctag,
				sopen: self.tagStrictOpen,
				text: buf,
				info: evInfo,
				parser: self,
				buffered: null,
				customParser: self.currentCustomParser
			};
			if (ctag) { ev.tag = self.decodeTag(ctag, ev); }
			if (cattr) { ev.attr = self.decodeAttr(cattr, ev); }
			if (buf) { ev.text = self.decodeText(buf, ev); }
			if (
				self.tagStrictOpen &&
				srcEv !== ev_text &&
				!(evInfo && (
					evInfo.strictOpenEvent ||
					evInfo.strictCloseEvent
				))
			) {
				ev.buffered = {
					pos: self.pos,
					line: self.line,
					column: self.column,
					endPos: self.endPos,
					endLine: self.endLine,
					endColumn: self.endColumn,
				};
				self.tagStrictEventBuf.push(ev);
			} else {
				eventFn(ev);
			}
		}
		function eventStartTag(id) {
			event(ev_startTag, id);
		}
		function eventTagName(id) {
			event(ev_tagName, id);
		}
		function eventEndTag(id, ev_custom) {
			var ev = ev_custom || ev_endTag;
			var sopen = self.tagStrictOpen;
			var evSkip = false;
			var evInfo = {
				strictOpenEvent: false,
				strictCloseEvent: false,
			};
			if (ev === ev_endTag) {
				if (sopen) {
					if (ctag.close && sopen === ctag.name) {
						evInfo.strictCloseEvent = true;
						event(ev_info, '380', evInfo);
						self.tagStrictEventBuf.forEach(function (ev) { return eventFn(ev); });
						event(ev_info, '390', evInfo);
					} else {
						evSkip = true;
						buf = self.tagStrictBuf + self.c;
						self.tagStrictBuf = '';
						self.tagStrictEventBuf = [];
						event(ev_text, '370', evInfo);
					}
				} else if (
					!ctag.close &&
					!ctag.selfClose &&
					self.isStrictTag(ctag.name)
				) {
					evInfo.strictOpenEvent = true;
					self.tagStrictOpen = ctag.name;
					self.tagStrictBuf = '';
					self.tagStrictEventBuf = [];
				}
			}
			buf = '';
			state = st_TEXT;
			if (!evSkip) { event(ev, id, evInfo); }
			self.currentTag = ctag = null;
			if (evInfo.strictCloseEvent) {
				self.tagStrictOpen = null;
				self.tagStrictBuf = null;
				self.tagStrictEventBuf = null;
			}
		}
		function eventTagAttr(id) {
			event(ev_tagAttribute, id);
			self.currentAttr = cattr = null;
		}
		function eventTagNameSlash(id) {
			ctag.name = tagNameSlash;
			eventTagName(id);
			tagNameSlash = void 0;
		}
		var self = this;
		text = text || '';
		var tlen = text.length;
		var eventFn = this.event;
		var state = this.state;
		var ctag = this.currentTag;
		var cattr = this.currentAttr;
		var buf = this.buffer;
		var pos = this.pos;
		var line = this.line;
		var column = this.column;
		var attrQuoteChar = this.attrQuoteChar;
		var tagBeforeClose = this.tagBeforeClose;
		var tagNameSlash = this.tagNameSlash;
		var ccp = this.currentCustomParser;
		for (var i = 0; i < tlen; i++) {
			var c = this.c = text[i];
			pos++;
			if (c === '\n') {
				line++;
				column = 0;
			} else {
				column++;
			}
			this.endPos = pos;
			this.endLine = line;
			this.endColumn = column;
			if (this.tagStrictOpen && c === '<') {
				if (ctag) {
					buf = this.tagStrictBuf;
					this.tagStrictBuf = '';
					this.currentTag = ctag = null;
					this.currentAttr = cattr = null;
					// console.error(
					// 	`TagStrictOpen:`,
					// 	this.tagStrictOpen,
					// 	state,
					// 	ctag,
					// 	cattr,
					// 	{ buf, c },
					// );
				}
				state = st_TEXT;
			}
			switch (state) {
				case st_TEXT:
					switch (c) {
						case '<':
							if (buf) {
								event(ev_text, '10');
								buf = '';
							}
							state = st_TAG_START;
							this.currentTag = ctag = {};
							eventStartTag('20');
							break;
						default:
							buf += c;
					}
					break;
				case st_TAG_START:
					switch (c) {
						case '?':
							if (buf) { ctag.startSpace = buf; }
							ctag.instruction = true;
							buf = '';
							state = st_INSTRUCTION;
							event(ev_startInstruction, '30');
							break;
						case '!':
							if (buf) { ctag.startSpace = buf; }
							// ctag.declaration = true;
							buf = '';
							state = st_DECLARATION_START;
							// event(ev_startDeclaration, '40');
							break;
						case '>':
							ctag.empty = true;
							if (buf) {
								if (ctag.close) {
									ctag.endSpace = buf;
								} else {
									ctag.startSpace = buf;
								}
								buf = '';
							}
							eventTagName('40');
							eventEndTag('50');
							break;
						case '/':
							ctag.close = true;
							if (buf) { ctag.startSpace = buf; }
							buf = '';
							break;
						default:
							if (reSpace.test(c)) {
								buf += c;
							} else {
								if (!buf) ; else if (ctag.close) {
									ctag.startCloseSpace = buf;
								} else {
									ctag.startSpace = buf;
								}
								state = st_TAG_NAME;
								buf = c;
							}
					}
					break;
				case st_INSTRUCTION:
					switch (c) {
						case '>':
							ctag.text = buf;
							eventEndTag('60', ev_endInstruction);
							break;
						default:
							buf += c;
					}
					break;
				case st_DECLARATION_START:
					switch (c) {
						case '>':
							ctag.declaration = true;
							ctag.text = buf;
							state = st_TEXT;
							event(ev_startDeclaration, '70');
							eventEndTag('80', ev_endDeclaration);
							break;
						case '-':
							if ('' === buf) {
								buf += c;
								break;
							} else if (buf === '-') {
								ctag.comment = true;
								buf = '';
								state = st_COMMENT;
								event(ev_startComment, '90');
								break;
							}
						default:
							buf += c;
							if (buf === strCdata) {
								ctag.cdata = true;
								buf = '';
								state = st_CDATA;
								event(ev_startCdata, '100');
							} else if (
								buf.length < strCdata.length &&
								strCdata.substr(0, buf.length) === buf
							) {
								break;
							} else {
								ctag.declaration = true;
								state = st_DECLARATION;
								event(ev_startDeclaration, '110');
								break;
							}
					}
					break;
				case st_DECLARATION:
					switch (c) {
						case '>':
							ctag.text = buf;
							eventEndTag('120', ev_endDeclaration);
							break;
						default:
							buf += c;
					}
					break;
				case st_COMMENT:
					switch (c) {
						case '>':
							if (getEndChars(3) === '-->') {
								ctag.textComment = buf.substr(0, buf.length - 2);
								buf = '';
								state = st_TEXT;
								event(ev_endComment, '130');
								break;
							}
						default:
							buf += c;
					}
					break;
				case st_CDATA:
					switch (c) {
						case '>':
							if (getEndChars(3) === ']]>') {
								ctag.textCdata = buf.substr(0, buf.length - 2);
								buf = '';
								state = st_TEXT;
								event(ev_endCdata, '140');
								break;
							}
						default:
							buf += c;
					}
					break;
				case st_TAG_NAME:
					switch (c) {
						case '>':
							if (tagBeforeClose) {
								ctag.name = tagBeforeClose;
								ctag.selfClose = true;
								this.tagBeforeClose = tagBeforeClose = void 0;
							} else {
								ctag.name = buf;
							}
							eventTagName('150');
							eventEndTag('160');
							break;
						case '/':
							if (ctag.close) {
								buf += c;
							} else {
								this.tagBeforeClose = tagBeforeClose = buf;
								buf = '';
							}
							break;
						default:
							if (reSpace.test(c)) {
								if (tagBeforeClose) {
									this.tagNameSlash = tagNameSlash = tagBeforeClose;
									this.tagBeforeClose = tagBeforeClose = void 0;
									state = st_TAG_NAMED_SELFCLOSE;
								} else {
									ctag.name = buf;
									state = st_TAG_NAMED;
									eventTagName('170');
								}
								buf = c;
							} else if (tagBeforeClose) {
								buf = tagBeforeClose+'/'+c;
								this.tagBeforeClose = tagBeforeClose = void 0;
							} else {
								buf += c;
							}
					}
					break;
				case st_TAG_NAMED:
					switch (c) {
						case '>':
							if (buf) { ctag.endSpace = buf; }
							eventEndTag('180');
							break;
						case '/':
							this.tagBeforeClose = tagBeforeClose = buf;
							state = st_TAG_NAMED_SELFCLOSE;
							buf = '';
							break;
						default:
							if (reSpace.test(c)) {
								buf += c;
							} else {
								state = st_ATTR_NAME;
								this.currentAttr = cattr = this.createAttr();
								if (buf) { cattr.startSpace = buf; }
								buf = c;
							}
					}
					break;
				case st_TAG_NAMED_SELFCLOSE:
					switch (c) {
						case '>':
							if (tagBeforeClose) { ctag.selfCloseSpace = tagBeforeClose; }
							if (buf) { ctag.endSpace = buf; }
							if (tagNameSlash) { eventTagNameSlash('190'); }
							ctag.selfClose = true;
							this.tagBeforeClose = tagBeforeClose = void 0;
							eventEndTag('200');
							break;
						case '/':
							if (tagNameSlash) {
								this.tagNameSlash = tagNameSlash += c;
								eventTagNameSlash('210');
							} else if (tagBeforeClose) {
								this.currentAttr = cattr = this.createAttr();
								cattr.startSpace = tagBeforeClose;
								cattr.name = c;
								eventTagAttr('220');
							}
							this.tagBeforeClose = tagBeforeClose = buf;
							buf = '';
							break;
						default:
							if (reSpace.test(c)) {
								buf += c;
							} else {
								if (tagNameSlash) {
									this.tagNameSlash = tagNameSlash += '/';
									eventTagNameSlash('230');
								}
								state = st_ATTR_NAME;
								this.currentAttr = cattr = this.createAttr();
								if (buf) { cattr.startSpace = buf; }
								buf = c;
							}
					}
					break;
				case st_ATTR_NAME:
					switch (c) {
						case '>':
							cattr.name = buf;
							eventTagAttr('240');
							eventEndTag('250');
							break;
						case '=':
							cattr.name = buf;
							state = st_ATTR_VALUE_START;
							buf = '';
							break;
						default:
							if (reSpace.test(c)) {
								cattr.name = buf;
								state = st_ATTR_NAMED;
								buf = c;
							} else {
								buf += c;
							}
					}
					break;
				case st_ATTR_NAMED:
					switch (c) {
						case '>':
							ctag.endSpace = buf;
							eventTagAttr('260');
							eventEndTag('270');
							break;
						case '=':
							if (buf) { cattr.eqSpace = buf; }
							cattr.eq = true;
							state = st_ATTR_VALUE_START;
							buf = '';
							break;
						default:
							if (reSpace.test(c)) {
								buf += c;
							} else {
								if (cattr) {
									eventTagAttr('280');
								}
								state = st_ATTR_NAME;
								this.currentAttr = cattr = this.createAttr();
								if (buf) { cattr.startSpace = buf; }
								buf = c;
							}
					}
					break;
				case st_ATTR_VALUE_START:
					switch (c) {
						case '>':
							if (buf) { ctag.endSpace = buf; }
							eventTagAttr('290');
							eventEndTag('300');
							break;
						case '"':
						case '\'':
							cattr.valueSpace = buf;
							this.attrQuoteChar = cattr.quotes = attrQuoteChar = c;
							state = st_ATTR_VALUE_QUOTED;
							buf = '';
							break;
						default:
							if (reSpace.test(c)) {
								buf += c;
							} else {
								if (buf) { cattr.valueSpace = buf; }
								ccp.attrValue = this.getCustomParser.attrValue(c, cattr);
								if (ccp.attrValue) {
									state = st_ATTR_VALUE_CUSTOM;
									buf = '';
								} else {
									state = st_ATTR_VALUE_RAW;
									buf = c;
								}
							}
					}
					break;
				case st_ATTR_VALUE_RAW:
					switch (c) {
						case '>':
							cattr.value = buf;
							eventTagAttr('310');
							eventEndTag('320');
							break;
						default:
							if (reSpace.test(c)) {
								cattr.value = buf;
								state = st_TAG_NAMED;
								eventTagAttr('330');
								buf = c;
							} else {
								buf += c;
							}
					}
					break;
				case st_ATTR_VALUE_QUOTED:
					switch (c) {
						case attrQuoteChar:
							this.attrQuoteChar = attrQuoteChar = void 0;
							cattr.value = buf;
							buf = '';
							state = st_TAG_NAMED;
							eventTagAttr('340');
							break;
						default:
							buf += c;
					}
					break;
				case st_ATTR_VALUE_CUSTOM:
					buf = ccp.attrValue.push(c, {attr: cattr, i: i, pos: pos, line: line, column: column});
					if (buf != void 0) {
						state = st_TAG_NAMED;
						ccp.attrValue = null;
						((assign = buf, cattr = assign.attr, i = assign.i, pos = assign.pos, line = assign.line, column = assign.column));
						buf = '';
						if (cattr) { eventTagAttr('350'); }
					}
					break;
			}
			this.pos = this.endPos;
			this.line = this.endLine;
			this.column = this.endColumn;
			if (this.tagStrictOpen && ctag) {
				this.tagStrictBuf += c;
			}
		}
		this.c = void 0;
		this.state = state;
		if (final && state === st_TEXT && buf) {
			event(ev_text, '360');
			buf = '';
		}
		this.buffer = buf;
	},
	end: function(text) {
		this.write(text, true);
	}
};

// import {arrayFrom, arrayConcat} from './collection';

function listToNode(list, adapter) {
	var node = adapter.initRoot();
	adapter.childrenSet(node, list);
	return node;
}

function treeRenderPlugin(sourceTree, sourceAdapter, ctx, plugin, targetAdapter, createTree, targetTree) {
	if (!ctx) { ctx = {}; }
	if (!targetAdapter) { targetAdapter = sourceAdapter; }

	var sourceArray = sourceAdapter.isChildren(sourceTree);
	if (sourceArray) { sourceTree = listToNode(sourceTree, sourceAdapter); }

	if (createTree) { targetTree = targetAdapter.initRoot(); }
	else if (!targetTree) { targetTree = sourceTree; }
	else if (targetAdapter.isChildren(targetTree)) {
		targetTree = listToNode(targetTree, targetAdapter);
	}

	var tc = sourceAdapter.childCount(sourceTree);
	var nodeInput, nodeOutput, arrayOutput, j, jOffset = 0;
	var parentBreadcrumb = ctx.getBreadcrumb;
	var getBreadcrumbLast = function() {
		return {
			// node: tree[j],
			index: j,
			indexOffset: jOffset,
			indexTarget: j+jOffset,
			node: sourceAdapter.childIndexGet(sourceTree, j),
			parent: sourceTree,
			parentTarget: targetTree,
		};
	};
	var getBreadcrumb = function() {
		var par = parentBreadcrumb && parentBreadcrumb();
		var last = getBreadcrumbLast();
		return par ? par.concat( [last]) : [last];
	};
	for (j = 0; j < tc; j++) {
		nodeInput = sourceAdapter.childIndexGet(sourceTree, j);
		ctx.getBreadcrumbLast = getBreadcrumbLast;
		ctx.getBreadcrumb = getBreadcrumb;
		nodeOutput = plugin(nodeInput, sourceAdapter, ctx, targetAdapter, createTree, targetTree);
		arrayOutput = targetAdapter.toArray(nodeOutput);
		var nc = arrayOutput.length - 1;
		targetAdapter.childSplice(targetTree, j+jOffset, 1, arrayOutput);
		if (targetTree === sourceTree) {
			j += nc, tc += nc;
		} else {
			jOffset += nc;
		}
	}
	ctx.getBreadcrumb = parentBreadcrumb;
	if (sourceArray) {
		targetTree = targetAdapter.childrenGet(targetTree);
	}
	return targetTree;
}

function treeRender(sourceTree, sourceAdapter, ctx, plugins) {
	var assign;

	var pc = plugins.length, i;
	for (i = 0; i < pc; i++) {
		var plugin = plugins[i];
		var targetAdapter, createTree, targetTree;
		if (!(plugin instanceof Function)) {
			((assign = plugin, plugin = assign.plugin, targetAdapter = assign.targetAdapter, createTree = assign.createTree, targetTree = assign.targetTree));
		}
		sourceTree = treeRenderPlugin(sourceTree, sourceAdapter, ctx, plugin, targetAdapter, createTree, targetTree);
		if (targetAdapter) { sourceAdapter = targetAdapter; }
	}
	return sourceTree;
}

function pluginListChildren(node, adapter, ctx, plugins) {
	if (!adapter.isText(node)) {
		var tree = adapter.childrenGet(node);
		if (tree) { adapter.childrenSet(node, treeRender(tree, adapter, ctx, plugins)); }
	}
	return node;
}

function pluginChildren(node, adapter, ctx, plugin, createTree) {
	if (!adapter.isText(node)) {
		var tree = adapter.childrenGet(node);
		if (tree) { adapter.childrenSet(node, treeRenderPlugin(tree, adapter, ctx, plugin, targetAdapter, createTree)); }
	}
	return node;
}

function pluginAttr(node, adapter, ctx, attrHandler, resultNode) {
	adapter.attrsEach(node, function (name, value, attr, index) {
		attrHandler(name, value, node, ctx, attr, index);
	});
	return resultNode ? resultNode(node, ctx) : node;
}

// plugin(nodeInput, sourceAdapter, ctx, targetAdapter, createTree);
function pluginConvertElement(node, adapter, ctx, targetAdapter, createTree) {
	var nodeTo;
	if (adapter.isText(node)) {
		nodeTo = targetAdapter.textNode(adapter.textValueGet(node));
	} else {
		nodeTo = targetAdapter.initName(adapter.nameGet(node));
		adapter.attrsEach(node, function (name, value, attr) {
			if (!attr) { attr = {name: name, value: value}; }
			targetAdapter.attrsAdd(nodeTo, attr);
		});
		// children = targetAdapter.childrenGet(nodeTo);
		treeRenderPlugin(node, adapter, ctx, pluginConvertElement, targetAdapter, createTree, nodeTo);
		// treeRenderPlugin(children, adapter, ctx, pluginConvertElement, targetAdapter, createTree);
		// treeRenderPlugin(sourceTree, sourceAdapter, ctx, plugin, targetAdapter, createTree) {
	}
	return nodeTo;
}

function adapterPluginConvertElement(targetAdapter) {
	return plugin;
	function plugin(nodeInput, sourceAdapter, ctx, _, createTree) {
		return pluginConvertElement(nodeInput, sourceAdapter, ctx, targetAdapter, createTree);
	}
}

var treerender = /*#__PURE__*/Object.freeze({
__proto__: null,
listToNode: listToNode,
treeRenderPlugin: treeRenderPlugin,
treeRender: treeRender,
pluginListChildren: pluginListChildren,
pluginChildren: pluginChildren,
pluginAttr: pluginAttr,
pluginConvertElement: pluginConvertElement,
adapterPluginConvertElement: adapterPluginConvertElement
});

var slice = Array.prototype.slice;

var events = {
	startTag: function(ev) {
		this.treeEvent('tagInit', null, ev);
	},
	startInstruction: function(ev) {
		this.scopeNewChild(TreeBuilder.TAG_INSTRUCTION, ev);
		this.treeEvent('tagInstructionStart', null, ev);
	},
	endInstruction: function(ev) {
		var cs = this.currentScope;
		var parent = cs;
		var tag = this.element.initInstruction(ev.tag.text);
		if (cs.tag !== TreeBuilder.TAG_INSTRUCTION) {
			var err = new TreeError(
				'Current open tag is not an instruction',
				120,
				this.getScopeState(ev, tag)
			);
			this.errors.push(err);
			this.treeEvent('error', err, ev);
		} else {
			cs.tag = tag;
			parent = cs.parentScope;
		}
		this.element.childElement(parent.tag, tag);
		this.treeEvent('tagInstructionEnd', null, ev);
		this.currentScope = parent;
	},
	startDeclaration: function(ev) {
		this.scopeNewChild(TreeBuilder.TAG_DECLARATION, ev);
		this.treeEvent('tagDeclarationStart', null, ev);
	},
	endDeclaration: function(ev) {
		var cs = this.currentScope;
		var parent = cs;
		var tag = this.element.initDeclaration(ev.tag.text);
		if (cs.tag !== TreeBuilder.TAG_DECLARATION) {
			var err = new TreeError(
				'Current open tag is not a declaration',
				121,
				this.getScopeState(ev, tag)
			);
			this.errors.push(err);
			this.treeEvent('error', err, ev);
		} else {
			cs.tag = tag;
			parent = cs.parentScope;
		}
		this.element.childElement(parent.tag, tag);
		this.treeEvent('tagDeclarationEnd', null, ev);
		this.currentScope = parent;
	},
	startComment: function(ev) {
		this.scopeNewChild(TreeBuilder.TAG_COMMENT, ev);
		this.treeEvent('tagCommentStart', null, ev);
	},
	endComment: function(ev) {
		var cs = this.currentScope;
		var parent = cs;
		var tag = this.element.initComment(ev.tag.textComment);
		if (cs.tag !== TreeBuilder.TAG_COMMENT) {
			var err = new TreeError(
				'Current open tag is not a comment',
				122,
				this.getScopeState(ev, tag)
			);
			this.errors.push(err);
			this.treeEvent('error', err, ev);
		} else {
			cs.tag = tag;
			parent = cs.parentScope;
		}
		this.element.childElement(parent.tag, tag);
		this.treeEvent('tagCommentEnd', null, ev);
		this.currentScope = parent;
	},
	tagName: function(ev) {
		if (ev.tag.close) {
			this.findAndCloseTag(ev);
		} else {
			this.openTag(ev);
		}
	},
	tagAttribute: function(ev) {
		this.element.attrsAdd(this.currentScope.tag, ev.attr, ev, this);
	},
	endTag: function(ev) {
		var breadcrumb = this.currentScope;
		var tagClose = ev.tag.close;
		var tagSelfClose = ev.tag.selfClose;
		if (!tagClose) {
			this.treeEvent('tagOpenEnd', null, ev);
			this.path.push(breadcrumb);
			this.element.childElement(breadcrumb.parentScope.tag, breadcrumb.tag, ev, this);
		}
		if (tagSelfClose) {
			this.findAndCloseTag(ev);
		}
		if (tagClose || tagSelfClose) {
			var tagOpen = this.closeTagMatch;
			// if tagOpen has no value, an error was already fired
			// in the function findAndCloseTag
			if (tagOpen) {
				var breadcrumbClose = tagOpen.match;
				this.treeEvent('tagCloseEnd', null, ev);
				this.path = this.path.slice(0, tagOpen.pathIndex);
				this.currentScope = breadcrumbClose.parentScope;
				this.closeTagMatch = null;
			}
		}
	},
	text: function(ev) {
		this.treeEvent('text', null, ev);
		this.element.childText(this.currentScope.tag, ev.text, ev, this);
	},
	info: function(ev) {
		this.treeEvent('info', null, ev);
	}
};

function noop(){}

function TreeError(message, code, state) {
	this.message = message;
	this.code = code;
	this.state = state;
	if (arguments.length > 2) { this.extras = slice.call(arguments, 2); }
}
TreeError.prototype = new Error;
TreeError.prototype.constructor = TreeError;
TreeError.prototype.name = 'TreeError';
TreeError.prototype.code = NaN;
TreeError.prototype.state = null;
TreeError.prototype.extras = null;

function TreeBuilder(opt) {
	if (opt instanceof Function) { opt = {event:opt}; }
	if (opt) { opt = ObjectAssign({}, TreeBuilder.optDefault, opt); }
	else { opt = ObjectAssign({}, TreeBuilder.optDefault); }
	this.eventFn = opt.event || noop;
	this.element = opt.element;
	this.tagVoidMap = opt.tagVoidMap;
	this.scopeNewChild(this.element.initRoot());
	this.root = this.currentScope;
	this.path = [];
	this.errors = [];
	this.parserEvent = this.parserEvent.bind(this);
}
TreeBuilder.optDefault = {
	element: null,
	tagVoidMap: null
};
TreeBuilder.TAG_INSTRUCTION = {tag:'#instruction'};
TreeBuilder.TAG_DECLARATION = {tag:'#declaration'};
TreeBuilder.TAG_COMMENT = {tag:'#comment'};
TreeBuilder.prototype = {
	constructor: TreeBuilder,
	eventFn: null,
	root: null,
	currentScope: null,
	closeTagMatch: null,
	path: null,
	errors: null,
	tagVoidMap: null,
	parserEvent: function(ev) {
		var handler = this.events[ev.name];
		if (handler) {
			handler.call(this, ev);
		}
	},
	getParserState: function(ref) {
		var parser = ref.parser;
		var id = ref.id;

		return {
			chr: parser.c,
			pos: parser.pos,
			line: parser.line,
			column: parser.column,
			endPos: parser.endPos,
			endLine: parser.endLine,
			endColumn: parser.endColumn,
			currentTag: parser.currentTag,
			currentAttr: parser.currentAttr,
			eventId: id
		};
	},
	getEventObject: function(name, error, event) {
		var cs = this.currentScope;
		return {
			name: name,
			error: error,
			event: event,
			tag: cs.tag,
			text: event.text,
			parent: cs.parentScope,
			path: this.path,
			tagClose: this.closeTagMatch,
			builder: this
		};
	},
	getScopeState: function(evEnd, tagEnd) {
		var cs = this.currentScope;
		return {
			tag: cs.tag,
			evStart: cs.evStart,
			evEnd: evEnd,
			tagEnd: tagEnd
		};
	},
	treeEvent: function(name, error, ev) {
		this.eventFn(this.getEventObject(name, error, ev));
	},
	scopeNewChild: function(child, ev) {
		var parent = this.currentScope;
		this.currentScope = {
			tag: child,
			evStart: ev,
			parentScope: parent
		};
	},
	openTag: function(ev) {
		this.scopeNewChild(this.element.initName(ev.tag.name), ev);
		this.treeEvent('tagOpenStart', null, ev);
	},
	unclosedTagChildren: function(tag) { // , index, ev
		var name = this.element.nameGet(tag);
		name = String(name).toLowerCase();
		if (this.tagVoidMap[name]) { return 0; }
	},	
	resolveUnclosedTags: function(unclosed, ev) {
		var uclen = unclosed.length;
		while (uclen) {
			var ucIndex = uclen - 1;
			var uc = unclosed[ucIndex];
			var ucTag = uc.tag;
			var ucEv = this.getEventObject('resolveUnclosedTags', null, ev);
			var ucCount = this.unclosedTagChildren(ucTag, ucIndex, ucEv);
			if (ucCount === void 0) { break; }
			var ucParent = uc.parentScope.tag;
			var ucSiblings = ucParent.children;
			var ucSelfPos = ucSiblings.indexOf(ucTag);
			if (-1 === ucSelfPos) {
				err = new TreeError(
					'Could not find child tag '+JSON.stringify(ucTag.name)+
					' in parent tag '+JSON.stringify(ucParent.name)+
					' list of '+ucSiblings.length+' children',
					103
				);
				this.errors.push(err);
				this.treeEvent('error', err, ev);
				break;
			} else {
				var ucRemCount = ucTag.children.length - ucCount;
				var ucRem = this.element.childSplice(ucTag, ucCount, ucRemCount, [], ev, this);
				this.element.childSplice(ucParent, ucSelfPos+1, 0, ucRem, ev, this);
				this.treeEvent('tagCloseStart', null, ev);
				this.treeEvent('tagCloseEnd', null, ev);
				unclosed.pop();
				uclen--;
				this.path.pop();
				this.currentScope = this.currentScope.parentScope;
			}
		}
		return unclosed;
	},
	findAndCloseTag: function(ev) {
		var err;
		var streamTag = ev.tag;
		var tagOpen = this.findTagOpen(streamTag, ev);
		this.closeTagMatch = tagOpen;
		if (tagOpen) {
			var breadcrumb = tagOpen.match;
			if (streamTag.selfClose) { breadcrumb.tag.selfClose = true; }
			var unclosed = this.resolveUnclosedTags(tagOpen.unclosedTags, ev);
			var uclen = unclosed.length;
			if (uclen) {
				err = new TreeError(
					(streamTag.selfClose ? 'Self closing tag ' : 'Close tag ')+
					JSON.stringify(streamTag.name)+' with '+uclen+' open tags',
					streamTag.selfClose ? 101 : 100
				);
				this.errors.push(err);
				this.treeEvent('unclosedTags', err, ev);
			}
			this.treeEvent('tagCloseStart', null, ev);
		} else {
			err = new TreeError(
				'Close tag '+JSON.stringify(streamTag.name)+' without opening tag',
				102
			);
			this.errors.push(err);
			this.treeEvent('unopenedTag', err, ev);
		}
	},
	findTagOpen: function(streamTag, ev) {
		var p = this.path;
		var match, i;
		for (i = p.length - 1; 0 <= i; i--) {
			var ptag = p[i];
			if (this.element.nameGet(ptag.tag, ev, this) === streamTag.name) {
				match = ptag;
				break;
			}
		}
		if (match) { return {
			match: match,
			pathIndex: i,
			unclosedTags: p.slice(i+1)
		}; }
	},
	events: events
};

TreeBuilder.TreeError = TreeError;

function defaultTestAdapter (x) {
	return x;
}
function defaultGroupListInit () {
	return [];
}
function defaultGroupInit (test) {
	return {
		test: test,
		matches: []
	};
}
function defaultGroupAdd (group, groupList) {
	groupList.push(group);
}
function defaultGroupGetCount(group) {
	return group.matches.length;
}
function defaultMatchAdd (result, item, group, test) {
	if (group.test !== test) {
		console.error(test, '!==', group.test);
		throw new Error('Object "test" in group is not same than "test" added from match');
	}
	group.matches.push({result: result, item: item});
}

function testList(testList, itemList, ref) {
	if ( ref === void 0 ) ref = {};
	var groupListInit = ref.groupListInit; if ( groupListInit === void 0 ) groupListInit = defaultGroupListInit;
	var groupInit = ref.groupInit; if ( groupInit === void 0 ) groupInit = defaultGroupInit;
	var groupAdd = ref.groupAdd; if ( groupAdd === void 0 ) groupAdd = defaultGroupAdd;
	var groupGetCount = ref.groupGetCount; if ( groupGetCount === void 0 ) groupGetCount = defaultGroupGetCount;
	var matchAdd = ref.matchAdd; if ( matchAdd === void 0 ) matchAdd = defaultMatchAdd;
	var testAdapter = ref.testAdapter; if ( testAdapter === void 0 ) testAdapter = defaultTestAdapter;

	var testSrc, test, item, active, testGroup, testResult;
	var activeTest, activeList, activeMatches, newAttempt;
	var active = {
		testList: testList.slice(),
		itemList: itemList.slice(),
		matches: groupListInit(),
		nextGroup: null,
		forked: false
	};
	var attemptsList = [active];
	var failed = [];
	var activeCount = 1;
	while (activeCount) {
		active = attemptsList[0];
		activeTest = active.testList;
		activeList = active.itemList;
		activeMatches = active.matches;
		if (!activeTest.length) {
			return result(0 == activeList.length);
		}
		testSrc = activeTest.shift();
		test = testAdapter(testSrc);
		testGroup = active.nextGroup;
		if (!testGroup) {
			testGroup = groupInit(test);
			groupAdd(testGroup, activeMatches);
		}
		active.nextGroup = null;
		if (0 == activeList.length) {
			if (groupGetCount(testGroup) < test.min) {
				failBranch();
			}
			continue;
		}
		if (groupGetCount(testGroup) === test.max) {
			continue;
		}
		if (groupGetCount(testGroup) >= test.min && !active.forked) {
			newAttempt = {
				testList: activeTest.slice(),
				itemList: activeList.slice(),
				matches: activeMatches.slice(),
				nextMatch: null,
				forked: false
			};
			attemptsList.splice(test.greedy ? 1 : 0, 0, newAttempt);
			activeCount++;
			activeTest.unshift(testSrc);
			active.nextGroup = testGroup;
			active.forked = true;
			newAttempt = void 0;
			continue;
		}
		active.forked = false;
		item = activeList.shift();
		testResult = test.test(item);
		if (testResult) {
			matchAdd(testResult, item, testGroup, test, testSrc);
			activeTest.unshift(testSrc);
			active.nextGroup = testGroup;
		} else {
			failBranch();
		}
	}
	return result(false);
	function failBranch() {
		failed.push(active);
		active = null;
		attemptsList.shift();
		activeCount--;
	}
	function result(success) {
		return {
			success: success,
			active: active,
			failed: failed,
			attemptsList: attemptsList
		};
	}
}

var STRING = 'string';
var OBJECT = 'object';

function getMatcherFrom(item, elAdapter, opt) {
	if (item instanceof TreeMatcher) { return item; }
	var tm = new TreeMatcher(elAdapter);
	if (STRING === typeof item) {
		tm.name(item, opt && opt.name);
	} else if (item instanceof Array) {
		if (null != item[0]) {
			tm.name(item[0], opt && opt.name);
		}
		if (item[1] instanceof Array) {
			tm.attrFromArray(item[1], opt && opt.attrs);
		}
		if (item[2] instanceof Array) {
			tm.path(item[2], opt && opt.path);
		}
	} else if (OBJECT === typeof item) {
		if (null != item.name) {
			tm.name(item.name, item.nameOpt || opt && opt.name);
		}
		if (item.attrs instanceof Array) {
			tm.attrFromArray(item.attrs, item.attrsOpt || opt && opt.attrs);
		}
		if (item.path instanceof Array) {
			tm.path(item.path, item.pathOpt || opt && opt.path);
		}
	}
	return tm;
}

function TreeMatcher(elAdapter) {
	this.elAdapter = elAdapter;
	this.clearRules();
}
TreeMatcher.from = getMatcherFrom;
TreeMatcher.prototype = {
	constructor: TreeMatcher,
	elAdapter: null,
	rulesName: null,
	rulesAttrs: null,
	rulesPath: null,
	defaultName: {
		repeatMin: 1,
		repeatMax: 1,
		repeatGreedy: false
	},
	defaultAttr: {
		repeatMin: 1,
		repeatMax: Infinity,
		repeatGreedy: false
	},
	reSpace: /^\s+|\s+(?=\s)|\s+$/g,
	reLimit: /^(.*)(?: <(?:([?*+])|(\d*)(,)?(\d*))(\?)?>)$/,
	reAttr: /^([^\s=]+)=(("?).*\3)$/,
	clearRules: function() {
		this.rulesName = [];
		this.rulesAttrs = [];
		this.rulesPath = [];
	},
	getStringOpt: function(str) {
		var mat = String(str).match(this.reLimit);
		var repeatMin, repeatMax, repeatGreedy;
		if (mat) {
			repeatGreedy = !mat[6];
			switch (mat[2]) {
				case '?': repeatMin = 0, repeatMax = 1; break;
				case '*': repeatMin = 0, repeatMax = Infinity; break;
				case '+': repeatMin = 1, repeatMax = Infinity; break;
				default:
					repeatMin = +mat[3] || 0;
					repeatMax = mat[4]
						? +mat[5] || Infinity
						: repeatMin;
			}
		}
		return {
			str: mat ? mat[1] : str,
			opt: {
				repeatMin: repeatMin,
				repeatMax: repeatMax,
				repeatGreedy: repeatGreedy
			}
		};
	},
	optExtend: function(target) {
		var arguments$1 = arguments;

		var tmin, tmax, tgre, source;
		var ac = arguments.length;
		if (STRING === typeof target) {
			target = this.getStringOpt(' '+target).opt;
		}
		for (var i = 1; i < ac; i++) {
			tmin = target.repeatMin;
			tmax = target.repeatMax;
			tgre = target.repeatGreedy;
			source = arguments$1[i];
			if (!source) { continue; }
			if (STRING === typeof source) {
				source = this.getStringOpt(' '+source).opt;
			}
			target.repeatMin = +tmin === tmin
				? tmin
				: source.repeatMin;
			target.repeatMax = +tmax === tmax
				? tmax
				: source.repeatMax;
			target.repeatGreedy = !!tgre === tgre
				? tgre
				: source.repeatGreedy;
			target.source = target.source || source.source;
		}
		return target;
	},
	strSpaceLower: function(str) {
		return STRING === typeof str
			? String(str).replace(this.reSpace, '').toLowerCase()
			: str;
	},
	strRaw: function(str) {
		return str;
	},
	getNormalize: function(normalize) {
		return normalize instanceof Function ? normalize : this.strRaw;
	},
	acceptAll: function() {
		return true;
	},
	methodAnd: {
		reduce: function(b) {
			return {
				result: b,
				_break: !b
			};
		}
	},
	methodOr: {
		reduce: function(b) {
			return {
				result: b,
				_break: b
			};
		}
	},
	methodAndList: {
		init: function() {
			return {
				and: [],
				not: [],
				success: true
			};
		},
		reduce: function(b, a, ref) {
			var rule = ref.rule;

			a[b?'and':'not'].push(rule);
			a.success = a.success && b;
			return {
				result: a,
				_break: !b
			};
		}
	},
	methodOrList: {
		init: function() {
			return {
				yes: [],
				not: [],
				success: false
			};
		},
		reduce: function(b, a, ref) {
			var itemResult = ref.itemResult;

			a[b?'yes':'not'].push(itemResult);
			a.success = b || a.success;
			return {
				result: a,
				_break: b
			};
		}
	},
	methodOrCount: {
		init: function(rules) {
			var count = [], rc = rules.length;
			for (var i = 0; i < rc; i++) { count[i] = 0; }
			return {
				rules: rules,
				count: count,
				nomatch: 0,
				success: false
			};
		},
		reduce: function(b, a, ref) {
			var ruleIndex = ref.ruleIndex;

			if (b) {
				a.count[ruleIndex] += 1;
			} else if (ruleIndex + 1 === a.rules.length) {
				a.nomatch += 1;
			}
			return {
				result: a,
				_break: b
			};
		},
		final: function(result) {
			if (0 != result.nomatch) {
				result.success = false;
				return result;
			}
			var rules = result.rules;
			var count = result.count;
			var rc = rules.length;
			var success = true;
			for (var i = 0; i < rc; i++) {
				success = rules[i].repeatMin <= count[i] && count[i] <= rules[i].repeatMax;
				if (!success) { break; }
			}
			result.success = success;
			return result;
		}
	},
	getItemSuccessDefault: function(itemResult) {
		return itemResult;
	},
	getItemSuccessOrder: function(itemResult) {
		return itemResult.success;
	},
	testValue: function(test, value) {
		if (test instanceof RegExp) {
			if (test.test(value)) { return true; }
		} else if (test instanceof Function) {
			if (test(value)) { return true; }
		} else {
			if (String(test) == value) { return true; }
		}
		return false;
	},
	preTest: function(test, normalize) {
		return STRING === typeof test
			? normalize(test)
			: null == test
			? this.acceptAll
			: test;
	},
	initRule: function(opt, def, test, getSuccess) {
		var m = {
			repeatMin: null,
			repeatMax: null,
			repeatGreedy: null,
			source: null,
			test: null,
			getSuccess: null
		};
		this.optExtend(m, opt, def);
		m.test = test;
		m.getSuccess = getSuccess || this.getItemSuccessDefault;
		return m;
	},
	testRuleItem: function(ruleList, item, testRule, method, rrUse) {
		// var rn = this.rulesName;
		var rc = ruleList.length;
		method = method || this.methodOr;
		var rr = rrUse || {
			result: null,
			_break: false
		};
		if (!rrUse && method.init) {
			rr.result = method.init(ruleList);
		}
		for (var i = 0; i < rc; i++) {
			var rule = ruleList[i];
			var itemResult = testRule(rule, item);
			var itemSuccess = rule.getSuccess(itemResult);
			rr = method.reduce(itemSuccess, rr.result, {
				first: 0 === i,
				rule: rule,
				item: item,
				itemResult: itemResult,
				ruleIndex: i
			});
			if (rr._break) { break; }
		}
		if (!rrUse && method.final) {
			rr.result = method.final(rr.result);
		}
		return rr.result;
	},
	testRuleItemList: function(ruleList, itemList, testRule, method) {
		var ic = itemList.length;
		var rrUse = {
			result: null,
			_break: false
		};
		if (method.init) {
			rrUse.result = method.init(ruleList);
		}
		for (var i = 0; i < ic; i++) {
			rrUse.result = this.testRuleItem(ruleList, itemList[i], testRule, method, rrUse);
		}
		if (method.final) {
			rrUse.result = method.final(rrUse.result);
		}
		return rrUse.result;
	},
	testRuleOrder: function(ruleList, itemList, testAdapter) {
		return testList(ruleList, itemList, { testAdapter: testAdapter });
	},
	name: function(testName, opt) {
		var normalize = this.getNormalize(opt && opt.normalizeNodeName);
		if ('*' === testName) { testName = this.acceptAll; }
		testName = this.preTest(testName, normalize);
		var self = this;
		var def = this.defaultName;
		var m = this.initRule(opt, def, function(nodeName) {
			return self.testValue(testName, normalize(nodeName));
		});
		this.rulesName.push(m);
	},
	testNameRule: function(rule, nodeName) {
		return rule.test(nodeName);
	},
	testNodeName: function(node, method) {
		var nodeName = this.elAdapter.nameGet(node);
		return this.testRuleItem(
			this.rulesName,
			nodeName,
			this.testNameRule,
			method
		);
	},
	attr: function(testAttr, opt) {
		var normName = this.getNormalize(opt && opt.normalizeAttrName);
		var normValue = this.getNormalize(opt && opt.normalizeAttrValue);
		if (STRING === typeof testAttr) {
			testAttr = this.getAttrTestFromString(testAttr);
			opt = this.optExtend(opt, testAttr.opt);
		} else if (testAttr instanceof Array) {
			opt = this.optExtend(opt, testAttr[2]);
			testAttr = { name: testAttr[0], value: testAttr[1] };
		}
		testAttr.name = this.preTest(testAttr.name, normName);
		testAttr.value = this.preTest(testAttr.value, normValue);
		var self = this;
		var def = this.defaultAttr;
		var m = this.initRule(opt, def, function(attr) {
			return (
				self.testValue(testAttr.name, normName(attr.name)) &&
				self.testValue(testAttr.value, normValue(attr.value))
			);
		});
		this.rulesAttrs.push(m);
	},
	getAttrTestFromString: function(item) {
		item = String(item || '');
		item = this.getStringOpt(item);
		var opt = item.opt;
		var mat = item.str.match(this.reAttr);
		var name, value, eq;
		if (mat) {
			name = mat[1];
			value = mat[2];
			if ('"' === mat[3]) { value = JSON.parse(value); }
		} else {
			eq = str.indexOf('=');
			name = -1 === eq ? str : str.substr(0, eq);
			value = -1 === eq ? null : str.substr(eq+1);
		}
		return { name: name, value: value, opt: opt };
	},
	attrFromArray: function(list, preOpt) {
		var c = list.length;
		for (var i = 0; i < c; i++) {
			var item = list[i];
			if (STRING === typeof item) {
				item = this.getAttrTestFromString(item);
			}
			if (item instanceof Array) {
				// item = item.slice();
				item[2] = this.optExtend(item[2] || {}, preOpt);
				this.attr({name: item[0], value: item[1]}, item[2]);
			} else if (OBJECT === typeof item) {
				item.opt = this.optExtend(item.opt || {}, preOpt);
				this.attr(item, item.opt);
			}
		}
	},
	nodeAttrsToArray: function(node) {
		var list = [];
		this.elAdapter.attrsEach(node, function(name, value) {
			list.push({name: name, value: value});
		});
		return list;
	},
	testAttrRule: function(rule, attr) {
		return rule.test(attr);
	},
	testAttrItem: function(attr) {
		return this.testRuleItem(this.rulesAttrs, attr, this.testAttrRule);
	},
	testNodeAttrs: function(node, method) {
		return this.testRuleItemList(
			this.rulesAttrs,
			this.nodeAttrsToArray(node),
			this.testAttrRule,
			method || this.methodOrCount
		);
	},
	testPathAdapter: function(m) {
		if (!m.rulesName) {
			console.error('rulesName is '+(typeof m.rulesName));
			console.error(m);
		}
		var rName = m.rulesName[0];
		return {
			min: rName.repeatMin,
			max: rName.repeatMax,
			greedy: rName.repeatGreedy,
			test: function(node) {
				return m.testNodeName(node)
					&& m.testNodeAttrs(node);
			}
		}
	},
	path: function(testPath, opt) {
		var tpc = testPath.length;
		for (var i = 0; i < tpc; i++) {
			testPath[i] = getMatcherFrom(testPath[i], this.elAdapter, opt);
		}
		var self = this;
		var m = this.initRule(null, null, function(path) {
			return self.testRuleOrder(testPath, path, self.testPathAdapter);
		}, this.getItemSuccessOrder);
		this.rulesPath.push(m);
	},
	testPathRule: function(rule, path) {
		return rule.test(path);
	},
	testPath: function(path) {
		return this.testRuleItem(
			this.rulesPath,
			path,
			this.testPathRule,
			this.methodOrList
		);
	},
	testAll: function(testNode, testPath) {
		var name, attr, path, success = false;
		name = this.testNodeName(testNode, this.methodOrList);
		if (name.success) {
			attr = this.testNodeAttrs(testNode);
			if (attr.success) {
				path = this.testPath(testPath);
				success = path.success;
			}
		}
		return { success: success, name: name, attr: attr, path: path };
	}
};

var echo = function (x) { return x; };

function Printer(opt) {
	if (opt) { opt = ObjectAssign({}, Printer.optDefault, opt); }
	else { opt = ObjectAssign({}, Printer.optDefault); }
	this.elAdapter = opt.elAdapter;
	this.indent = opt.indent;
	this.encodeString = opt.encodeString || echo;
	this.encodeText = opt.encodeText || this.encodeString;
	this.encodeTagName = opt.encodeTagName || this.encodeString;
	this.encodeAttrName = opt.encodeAttrName || this.encodeString;
	this.encodeAttrValue = opt.encodeAttrValue || this.encodeString;
	this.selfCloseString = opt.selfCloseString || this.selfCloseString;
	this.newLine = opt.newLine || this.newLine;
	this.tagVoidMap = opt.tagVoidMap;
	this.tagStrictMap = opt.tagStrictMap;
	this.rootStrict = opt.rootStrict;
}

Printer.optDefault = {
	tagVoidMap: null,
	tagStrictMap: null,
	rootStrict: false,
};

Printer.prototype = {
	constructor: Printer,
	elAdapter: null,
	indent: null,
	encodeString: null,
	encodeText: null,
	encodeTagName: null,
	encodeAttrName: null,
	encodeAttrValue: null,
	selfCloseString: ' /',
	newLine: '\n',
	tagVoidMap: null,
	tagStrictMap: null,
	rootStrict: false,
	log: function() {},
	repeat: function(n, c) {
		var s = '';
		while (n > 0) {
			s += c;
			n--;
		}
		return s;
	},
	printIndent: function(level) {
		var indent = this.indent;
		if (indent instanceof Function) {
			return indent(level);
		}
		if (null == indent) {
			indent = '\t';
		} else if (+indent === indent) {
			indent = this.repeat(indent, ' ');
		}
		return this.repeat(level, indent);
	},
	isVoidTag: function(node) {
		var name = this.elAdapter.nameGet(node);
		name = String(name || '').toLowerCase();
		return this.tagVoidMap[name];
	},
	isStrictTag: function(node) {
		var name = this.elAdapter.nameGet(node);
		name = String(name || '').toLowerCase();
		return this.tagStrictMap[name];
	},
	isStrictPath: function(path) {
		var plen = path.length;
		var last = plen > 0 && path[plen - 1];
		return last ? this.isStrictTag(last) : this.rootStrict;
	},
	printAttr: function(name, value) {
		var attr = this.encodeAttrName(name);
		if (null != value) {
			attr += '="'+this.encodeAttrValue(value)+'"';
		}
		return attr;
	},
	printTagOpen: function(node, isSelfClose) {
		var self = this;
		var name = this.elAdapter.nameGet(node);
		var tag = '<' + this.encodeTagName(name);
		this.elAdapter.attrsEach(node, function(name, value) {
			tag += ' '+self.printAttr(name, value);
		});
		tag += (isSelfClose ? this.selfCloseString : '') + '>';
		return tag;
	},
	printTagClose: function(node) {
		var name = this.elAdapter.nameGet(node);
		var tag = '</' + this.encodeTagName(name) + '>';
		return tag;
	},
	printTagChildren: function(node, level, path) {
		return this.print(this.elAdapter.childrenGet(node), level, path);
	},
	printTagChildrenAsync: function(node, level, path, cbPrint) {
		return this.printAsync(this.elAdapter.childrenGet(node), level, path, cbPrint);
	},
	printTag: function(node, level, path) {
		var nl = this.newLine;
		var nc = this.elAdapter.childCount(node);
		var sc = 0 == nc && this.isVoidTag(node);
		var st = this.isStrictTag(node);
		var out = this.printIndent(level);
		out += this.printTagOpen(node, sc);
		if (nc > 0) {
			out += st ? '' : nl;
			out += this.printTagChildren(node, level+1, path.concat([node]));
			out += st ? '' : this.printIndent(level);
		}
		if (!sc) {
			out += this.printTagClose(node);
		}
		out += nl;
		return out;
	},
	printTagAsync: function(node, level, path, cbPrint) {
		var nc = this.elAdapter.childCount(node);
		if (nc) {
			this.printTagChildrenAsync(node, level+1, path.concat([node]), cbChildren.bind(this));
		} else {
			cbChildren.call(this);
		}
		function cbChildren(err, children) {
			var nl = this.newLine;
			var sc = 0 == nc && this.isVoidTag(node);
			var out = this.printIndent(level);
			out += this.printTagOpen(node, sc);
			if (children) {
				out += nl + children + this.printIndent(level);
			}
			if (!sc) {
				out += this.printTagClose(node);
			}
			out += nl;
			cbPrint(err, out);
		}
	},
	textSplitLines: function(text) {
		return text.split(/\r\n|\n|\r/g);
	},
	textTrim: function(text) {
		return text.replace(/^[ \t\r\n]+|[ \t]+(?=[ \t])|[ \t\r\n]+$/g, '');
	},
	printTextLine: function(text, level) {
		var trim = this.textTrim(text);
		return trim ? this.printIndent(level) + trim : trim;
	},
	printText: function(ftext, level, path) {
		var strict = this.isStrictPath(path);
		if (strict) { return ftext; }
		var text = this.textTrim(ftext);
		text = this.textSplitLines(text);
		var c = text.length;
		var nl = this.newLine;
		var s = '';
		for (var i = 0; i < c; i++) {
			var l = this.printTextLine(text[i], level);
			s += l ? l + nl : '';
		}
		return s;
	},
	printTextAsync: function(ftext, level, path, cbPrint) {
		return cbPrint(null, this.printText(ftext, level));
	},
	printComment: function(ftext, level) {
		return this.printIndent(level) +
			'<!--' + ftext + '-->\n';
	},
	printCommentAsync: function(ftext, level, path, cbPrint) {
		return cbPrint(null, this.printComment(ftext, level));
	},
	printDeclaration: function(ftext, level) {
		return this.printIndent(level) +
			'<!' + ftext + '>\n';
	},
	printDeclarationAsync: function(ftext, level, path, cbPrint) {
		return cbPrint(null, this.printDeclaration(ftext, level));
	},
	printInstruction: function(ftext, level) {
		return this.printIndent(level) +
			'<?' + ftext + '>\n';
	},
	printInstructionAsync: function(ftext, level, path, cbPrint) {
		return cbPrint(null, this.printInstruction(ftext, level));
	},
	print: function(tree, level, path) {
		var tc = tree.length;
		var out = '';
		path = path || [];
		for (var i = 0; i < tc; i++) {
			var node = tree[i];
			if (this.elAdapter.isText(node)) {
				out += this.printText(this.elAdapter.textValueGet(node), level, path);
			} else if (this.elAdapter.isComment(node)) {
				out += this.printComment(this.elAdapter.textValueGet(node), level, path);
			} else if (this.elAdapter.isDeclaration(node)) {
				out += this.printDeclaration(this.elAdapter.textValueGet(node), level, path);
			} else if (this.elAdapter.isInstruction(node)) {
				out += this.printInstruction(this.elAdapter.textValueGet(node), level, path);
			} else {
				out += this.printTag(node, level, path);
			}
		}
		return out;
	},
	printAsync: function(tree, level, path, cbPrint) {
		var out = '';
		path = path || [];
		cbNext = cbNext.bind(this);
		return runNext.call(this);
		function runNext() {
			if (tree.length) {
				var node = tree.shift();
				if (this.elAdapter.isText(node)) {
					this.printTextAsync(this.elAdapter.textValueGet(node), level, path, cbNext);
				} else if (this.elAdapter.isComment(node)) {
					this.printCommentAsync(this.elAdapter.textValueGet(node), level, path, cbNext);
				} else if (this.elAdapter.isDeclaration(node)) {
					this.printDeclarationAsync(this.elAdapter.textValueGet(node), level, path, cbNext);
				} else if (this.elAdapter.isInstruction(node)) {
					this.printInstructionAsync(this.elAdapter.textValueGet(node), level, path, cbNext);
				} else {
					this.printTagAsync(node, level, path, cbNext);
				}
			} else {
				cbPrint(null, out);
			}
		}
		function cbNext(err, printed) {
			if (err) { return cbPrint(err, out); }
			out += printed;
			runNext.call(this);
		}
	}
};

function arrayFrom(val) {
	var slice = Array.prototype.slice;
	switch (true) {
		case val instanceof Array: return val;
		case void 0 == val: return [];
		case isCollection(val): return slice.call(val);
		default: return [val];
	}
}

function arrayConcat(val) {
	switch (true) {
		case val instanceof Array: return val;
		case void 0 == val: return [];
		default: return [val];
	}
}

function isCollection(obj) {
	var propEnum = Object.prototype.propertyIsEnumerable;
	var len, intLen;
	return obj &&
		'length' in obj &&
		!propEnum.call(obj, 'length') &&
		(len = obj.length,
			intLen = parseInt(len),
			intLen === len && intLen >= 0);
}

function elementDefault (ref) {
	if ( ref === void 0 ) ref = {};
	var keyName = ref.keyName; if ( keyName === void 0 ) keyName = 'name';
	var keyAttrs = ref.keyAttrs; if ( keyAttrs === void 0 ) keyAttrs = 'attrs';
	var keyChildren = ref.keyChildren; if ( keyChildren === void 0 ) keyChildren = 'children';
	var keyText = ref.keyText; if ( keyText === void 0 ) keyText = 'text';
	var textName = ref.textName; if ( textName === void 0 ) textName = '#text';
	var rootName = ref.rootName; if ( rootName === void 0 ) rootName = '#document-fragment';
	var commentName = ref.commentName; if ( commentName === void 0 ) commentName = '#comment';
	var declarationName = ref.declarationName; if ( declarationName === void 0 ) declarationName = '#declaration';
	var instructionName = ref.instructionName; if ( instructionName === void 0 ) instructionName = '#instruction';

	var initName = function (name) {
		var obj;

		return (( obj = {}, obj[keyName] = name, obj[keyAttrs] = [], obj[keyChildren] = [], obj[keyText] = null, obj ));
	};
	var child = function (el, child) { return void el[keyChildren].push(child); };
	var textNode = function (text) {
		var obj;

		return (( obj = {}, obj[keyName] = textName, obj[keyText] = text, obj ));
	};
	return {
		isText: function (el) { return textName === el[keyName]; },
		isFragment: function (el) { return el[keyName] === rootName; },
		isComment: function (el) { return el[keyName] === commentName; },
		isDeclaration: function (el) { return el[keyName] === declarationName; },
		isInstruction: function (el) { return el[keyName] === instructionName; },
		isChildren: function (ch) { return ch instanceof Array; },
		initRoot: function () { return initName(rootName); },
		initName: initName,
		initComment: function (text, el) {
			if ( text === void 0 ) text = '';

			return (el = initName(commentName), el[keyText] = text, el);
	},
		initDeclaration: function (text, el) {
			if ( text === void 0 ) text = '';

			return (el = initName(declarationName), el[keyText] = text, el);
	},
		initInstruction: function (text, el) {
			if ( text === void 0 ) text = '';

			return (el = initName(instructionName), el[keyText] = text, el);
	},
		nameGet: function (el) { return el[keyName]; },
		textNode: textNode,
		textValueGet: function (el) { return el[keyText]; },
		textValueSet: function (el, text) { return el[keyText] = text; },
		attrsAdd: function (el, attr) { return void el[keyAttrs].push(attr); },
		attrsEach: function (el, handler) {
			var list = el[keyAttrs];
			if (!list) { console.log('ElementDefault attrs not found', el, keyAttrs); }
			var count = list && list.length || 0;
			var ctx = {
				_break: 1 << 0,
				_remove: 1 << 1
			};
			for (var i = 0; i < count; i++) {
				var a = list[i];
				var ret = handler.call(ctx, a.name, a.value, a, i);
				if (ret & ctx._remove) { list.splice(i, 1), i--, count--; }
				if (ret & ctx._break) { break; }
			}
		},
		childElement: child,
		childText: function (el, text) { return child(el, textNode(text)); },
		childCount: function (el) { return el[keyChildren].length; },
		childIndexGet: function (el, index) { return el[keyChildren][index]; },
		childSplice: function (el, index, remove, add) {
			var ref;

			return (ref = el[keyChildren]).splice.apply(ref, [ index, remove ].concat( (add || []) ));
	},
		childrenGet: function (el) { return el[keyChildren]; },
		childrenSet: function (el, children) { return el[keyChildren] = children; },
		toArray: arrayConcat,
	};
}

function dom (apiDom) {
	var child = function (el, child) { return void el.appendChild(child); };
	var childCount = function (el) { return el.childNodes.length; };
	var childIndexGet = function (el, index) { return el.childNodes[index]; };
	var childSplice = function (el, index, remove, add) {
		//splice.apply(el.childNodes, [index, remove, ...(add || [])]);
		var ec = childCount(el);
		while (remove > 0 && ec > index) {
			el.removeChild(childIndexGet(el, index));
			remove--, ec--;
		}
		var i, c = add.length;
		var temp = []; // Live NodeList changes as we reassign nodes
		for (i = 0; i < c; i++) {
			temp.push(add[i]);
		}
		add = temp;
		for (i = 0; i < c; i++) {
			el.insertBefore(add[i], childIndexGet(el, index + i));
		}
	};
	var textNode = function (text) { return apiDom.createTextNode(text); };
	var isFragment = function (el) { return el.nodeType === apiDom.DOCUMENT_FRAGMENT_NODE; };
	var isComment = function (el) { return el.nodeType === apiDom.COMMENT_NODE; };
	var isChildren = function (ch) { return ch && ch.constructor === apiDom.childNodes.constructor; };
	// const splice = Array.prototype.splice;
	return {
		isText: function (el) { return el.nodeType === apiDom.TEXT_NODE; },
		isFragment: isFragment,
		isComment: isComment,
		isChildren: isChildren,
		initRoot: function () { return apiDom.createDocumentFragment(); },
		initName: function (name) { return apiDom.createElement(name); },
		initComment: function (text) {
			if ( text === void 0 ) text = '';

			return apiDom.createComment(text);
	},
		nameGet: function (el) { return el.nodeName; },
		textNode: textNode,
		textValueGet: function (el) { return el.nodeValue; },
		textValueSet: function (el, text) { return void (el.nodeValue = text); },
		attrsAdd: function (el, attr) { return el.setAttribute(attr.name, attr.value); },
		attrsEach: function (el, handler) {
			var list = el.attributes;
			if (!list) { console.log('ElementDom attrs not found', el); }
			var ctx = {
				_break: 1 << 0,
				_remove: 1 << 1
			};
			var count = list.length;
			for (var i = 0; i < count; i++) {
				var a = list[i];
				var ret = handler.call(ctx, a.name, a.value, a, i);
				if (ret & ctx._remove) { el.removeAttribute(a.name); }
				if (ret & ctx._break) { break; }
			}
		},
		childElement: child,
		childText: function (el, text) { return child(el, textNode(text)); },
		childCount: childCount,
		childIndexGet: function (el, index) { return el.childNodes[index]; },
		childSplice: childSplice,
		childrenGet: function (el) { return el.childNodes; },
		childrenSet: function (el, children) { return childSplice(el, 0, childCount(el), children); },
		toArray: function (el) {
			return !el ? [] :
				isFragment(el) ? arrayFrom(el.childNodes) :
				el.constructor === apiDom.childNodes.constructor ? arrayFrom(el) :
				arrayConcat(el);
		}
	};
}

var initName = function (name) { return ({
	_name: name,
	_class: null,
	_id: null,
	sel: name,
	data: {},
	text: void 0, // Snabbdom requires undefined
	children: []
}); };
var textNode = function (text) { return ({text:text}); };
var child = function (el, child) { return void el.children.push(child); };
var attrIdSel = function (attr, el) {
	if (attr.name === 'id') {
		el._id = attr.value;
		attrUpdateSel(el);
		return true;
	}
};
var attrClassSel = function (attr, el) {
	if (attr.name === 'class') {
		el._class = attr.value;
		attrUpdateSel(el);
		return true;
	}
};
var attrUpdateSel = function (el) {
	var c = el._class;
	var i = el._id;
	el.sel = el._name +
		(i ? '#' + i : '') +
		(c ? '.' + c.replace(/\s+/g,'.') : '');
};
var attrObject = function (attr, el, key) {
	var ea = el.data[key];
	if (!ea) { el.data[key] = ea = {}; }
	ea[attr.name] = attr.value;
	return true;
};
var attrHandlerDefault = function (el, attr) {
	return attrIdSel(attr, el)
		|| attrClassSel(attr, el)
		|| attrObject(attr, el, 'attrs');
};
var attrFn = {
	attrIdSel: attrIdSel,
	attrClassSel: attrClassSel,
	attrUpdateSel: attrUpdateSel,
	attrObject: attrObject,
	attrHandlerDefault: attrHandlerDefault
};
var refFragment = {};
var nameComment = '!';

function snabbdom (ref) {
	if ( ref === void 0 ) ref = {};
	var attrHandler = ref.attrHandler;

	attrHandler = (attrHandler || attrHandlerDefault).bind(attrFn);
	return {
		isText: function (el) { return !el.sel && 'string' === typeof el.text; },
		isFragment: function (el) { return el._name === refFragment; },
		isComment: function (el) { return el._name === nameComment; },
		isChildren: function (ch) { return ch instanceof Array; },
		initRoot: function () { return initName(refFragment); },
		initName: initName,
		initComment: function (text) {
		if ( text === void 0 ) text = '';
var el = initName(nameComment); el.text = text; return el},
		nameGet: function (el) { return el._name; },
		textNode: textNode,
		textValueGet: function (el) { return el.text; },
		textValueSet: function (el, text) { return void (el.text = text); },
		attrsAdd: attrHandler,
		attrsEach: function (el, handler) {
			if (el && !el.data) { console.log('ElementSnabbdom data not found', el); }
			var map = el.data.attrs;
			var hop = Object.prototype.hasOwnProperty;
			if (!map) { return; }
			var ctx = {
				_break: 1 << 0,
				_remove: 1 << 1
			};
			var keys = [], k;
			for (k in map) { if (hop.call(map, k)) { keys.push(k); } }
			for (var i = 0, count = keys.length; i < count; i++) {
				k = keys[i];
				var ret = handler.call(ctx, k, map[k], null, i);
				if (ret & ctx._remove) { delete map[k]; }
				if (ret & ctx._break) { break; }
			}
		},
		childElement: child,
		childText: function (el, text) { return child(el, textNode(text)); },
		childCount: function (el) { return el.children.length; },
		childIndexGet: function (el, index) { return el.children[index]; },
		childSplice: function (el, index, remove, add) {
			var ref;

			return (ref = el.children).splice.apply(ref, [ index, remove ].concat( (add || []) ));
	},
		childrenGet: function (el) { return el.children; },
		childrenSet: function (el, children) { return el.children = children; },
		toArray: arrayConcat,
	};
}

var htmlVoidTagMap = {
	area:true,// clickable area inside an image map.
	base:true,// new root or base path for relative paths.
	basefont:true,
	bgsound:true,
	br:true,// single line break in a paragraph.
	col:true,
	command:true,
	embed:true,// embeds things like flash.
	// frame:true,
	hr:true,// change in type of content.
	// iframe:true,// frame inside the page. NOT A VOID TAG
	image:true,
	img:true,// image.
	input:true,// input field.
	isindex:true,
	keygen:true,
	link:true,// reference an external resource.
	menuitem:true,
	meta:true,// metadata about the page.
	nextid:true,
	param:true,// parameter for an object.
	source:true,// include media for video or audio.
	track:true,// text track for video or audio.
	wbr:true// word break opportunity
};

var htmlStrictTagMap = {
	script:true,
};

function getParser(elAdapter, tagVoidMap, unclosedTagChildren) {
	elAdapter = elAdapter || elementDefault();
	tagVoidMap = tagVoidMap || htmlVoidTagMap;
	var tb = new TreeBuilder({
		element: elAdapter,
		tagVoidMap: tagVoidMap
	});
	if (unclosedTagChildren) {
		tb.unclosedTagChildren = unclosedTagChildren;
	}
	var xp = new XMLParser(tb.parserEvent.bind(tb));

	return {
		write: xp.write.bind(xp),
		end: xp.end.bind(xp),
		getResult: function(ref) {
			if ( ref === void 0 ) ref = {};
			var asNode = ref.asNode;

			var error = tb.errors;
			var tree = asNode ? tb.root.tag : elAdapter.childrenGet(tb.root.tag);
			if (error instanceof Array && 0 === error.length) { error = null; }
			return {
				error: error,
				tree: tree,
				elAdapter: elAdapter,
				builder: tb,
				parser: xp
			};
		},
		elAdapter: elAdapter
	};
}

function printTreeSync(ref) {
	var tree = ref.tree;
	var elAdapter = ref.elAdapter;
	var customPrintTag = ref.customPrintTag;
	var path = ref.path;
	var level = ref.level;

	var printer = new Printer();
	printer.elAdapter = elAdapter;
	if (customPrintTag instanceof Function) {
		var defaultPrintTag = printer.printTag.bind(printer);
		printer.printTag = function(node, level, path) {
			return customPrintTag(node, level, path, defaultPrintTag, printer);
		};
	}
	return printer.print(tree, level || 0, path);
}

function printTreeAsync(ref) {
	var tree = ref.tree;
	var elAdapter = ref.elAdapter;
	var customPrintTag = ref.customPrintTag;
	var path = ref.path;
	var level = ref.level;
	var callback = ref.callback;

	var printer = new Printer();
	printer.elAdapter = elAdapter;
	if (customPrintTag instanceof Function) {
		var defaultPrintTag = printer.printTagAsync.bind(printer);
		printer.printTagAsync = function(node, level, path, cbTag) {
			return customPrintTag(node, level, path, defaultPrintTag, printer, cbTag);
		};
	}
	return printer.printAsync(tree, level || 0, path, callback);
}

function parseString(str, elAdapter) {
	var parser = getParser(elAdapter);
	parser.write(str);
	parser.end();
	return parser.getResult();
}

function prepare(rep, level, path, elAdapter, transformName) {
	var assign;

	var name = rep.name;
	var tree = rep.tree;
	var text = rep.text;
	var noFormat = rep.noFormat;
	var error;
	if (null == tree && !noFormat) {
		rep = parseString(String(text || ''), elAdapter);
		((assign = rep, tree = assign.tree, error = assign.error));
		if (error) {
			error = { name: name || transformName, error: error };
		}
	}
	if (null != tree) {
		level = null == rep.indent ? level :
			+rep.indent === rep.indent ? rep.indent :
			rep.indent instanceof Function ? rep.indent(level) :
			level;
		text = printTreeSync({tree: tree, elAdapter: elAdapter, path: path, level: level});
	} else if (!noFormat) {
		text =
			printer.printIndent(level) +
			text +
			printer.newLine;
	}
	return {error: error, text: text};
}

function apply(rep, level, path, elAdapter) {
	var errors = [], text;
	if (
		null != rep.full ||
		null != rep.fullSrc
	) {
		text = prepareRep( rep.before ) +
			( rep.full
				? prepareRep( rep.full )
				: rep.fullSrc || ''
			) +
			prepareRep( rep.after );
	} else {
		text =
			prepareRep( rep.before ) +
			rep.open + 
			prepareRep( rep.prepend, 1 ) +
			( rep.children
				? prepareRep( rep.children, 1 )
				: rep.childrenSrc || ''
			) +
			prepareRep( rep.append, 1 ) +
			rep.close +
			prepareRep( rep.after );
	}
	return {errors: errors, text: text};
	function prepareRep(trep, add) {
		if (trep) {
			trep = prepare(trep, level + (add || 0), path, elAdapter, rep.name);
			if (trep.error) { errors.push(trep.error); }
			return String(trep.text || '');
		} else {
			return '';
		}
	}
}

function async(ref) {
	var tree = ref.tree;
	var elAdapter = ref.elAdapter;
	var transform = ref.transform;
	var callback = ref.callback;
	var level = ref.level;

	var repErrors = [];
	return printTreeAsync({tree: tree, elAdapter: elAdapter, customPrintTag: customPrintTag, level: level, callback: cbPrintTree});
	function addErrors(err) {
		if (err instanceof Array) {
			repErrors = repErrors.concat(err);
		} else if (err) {
			repErrors.push(err);
		}
	}
	function cbPrintTree(err, page) {
		addErrors(err);
		if (!repErrors.length) { repErrors = null; }
		callback(repErrors, page);
	}
	function customPrintTag(node, level, path, printTagAsync, printer, cbTag) {
		return transform({node: node, path: path, level: level, elAdapter: elAdapter, printer: printer, callback: cbTransform});
		function cbTransform(err, rep) {
			if (rep) {
				addErrors(err);
				if (
					null == rep.full && (
					null != rep.append ||
					null != rep.prepend ||
					null != rep.children )
				) {
					rep.open =
						printer.printIndent(level) +
						printer.printTagOpen(node) +
						( null == rep.children || !rep.noFormat ? printer.newLine : '' );
					rep.close = 
						( null == rep.children || !rep.noFormat ? printer.printIndent(level) : '' ) +
						printer.printTagClose(node) +
						printer.newLine;
					if (null == rep.children) {
						return printer.printTagChildrenAsync(node, level+1, path.concat([node]), function(errPrint, nodeChildren) {
							addErrors(errPrint);
							rep.childrenSrc = nodeChildren;
							return withContainer(rep);
						});
					} else {
						return withContainer(rep);
					}
				} else if (
					null == rep.full
				) {
					return printTagAsync(node, level, path, function (errPrint, nodeFull) {
						addErrors(errPrint);
						rep.fullSrc = nodeFull;
						return withContainer(rep);
					});
				}
				return withContainer(rep);
			} else if (err) {
				return cbTag(err, '');
			} else {
				return printTagAsync(node, level, path, cbTag);
			}
		}
		function withContainer(rep) {
			var ref = apply(rep, level, path, elAdapter);
			var errors = ref.errors;
			var text = ref.text;
			addErrors(errors);
			cbTag(null, text);
		}
	}
}

function asyncMatcher(elAdapter) {
	var rules = [];
	var api = {
		addRule: function(opt) {
			opt.matcher = TreeMatcher.from(opt.matcher, elAdapter, opt.opt);
			rules.push(opt);
		},
		clear: function() {
			rules = [];
		},
		onTest: function() {},
		onTestRule: function() {},
		isSuccess: function(result) {
			return result.success;
		},
		transform: function(opt) {
			var node = opt.node;
			var path = opt.path;
			var rc = rules.length;
			api.onTest(opt);
			for (var i = 0; i < rc; i++) {
				var isSuccess = rules[i].isSuccess || api.isSuccess;
				var result = rules[i].matcher.testAll(node, path);
				var success = isSuccess(result);
				api.onTestRule(result, success, rules[i], opt);
				if (success) {
					return rules[i].callback(opt);
				}
			}
			opt.callback();
		}
	};
	return api;
}

var printertransform = /*#__PURE__*/Object.freeze({
__proto__: null,
printTreeSync: printTreeSync,
printTreeAsync: printTreeAsync,
prepare: prepare,
apply: apply,
async: async,
asyncMatcher: asyncMatcher
});

function getPosStart(parser) {
	return {
		byte: parser.pos,
		line: parser.line,
		column: parser.column
	};
}
function getPosEnd(parser) {
	return {
		byte: parser.endPos,
		line: parser.endLine,
		column: parser.endColumn
	};
}

var statsEventMap = {
	tagInit: function statsTagInit(ev) {
		this.statsLastTagInit = {
			tag: ev.event.tag,
			pos: getPosStart(ev.event.parser)
		};
	},
	tagOpenStart: function statsTagOpenStart(ev) {
		var s = ev.tag.stats;
		s || (s = ev.tag.stats = {});
		var sl = this.statsLastTagInit;
		var streamTag = ev.event.tag;
		s.openSource = streamTag;
		s.openStart = (sl && sl.tag === streamTag && sl.pos) || getPosStart(ev.event.parser);
	},
	tagOpenEnd: function statsTagOpenEnd(ev) {
		var s = ev.tag.stats;
		s || (s = ev.tag.stats = {});
		s.openEnd = getPosEnd(ev.event.parser);
		this.statsLastTagInit = null;
	},
	tagCloseStart: function statsTagCloseStart(ev) {
		var s = ev.tag.stats;
		s || (s = ev.tag.stats = {});
		var sl = this.statsLastTagInit;
		var streamTag = ev.event.tag;
		s.closeSource = streamTag;
		s.closeStart = (sl && sl.tag === streamTag && sl.pos) || getPosStart(ev.event.parser);
	},
	tagCloseEnd: function statsTagCloseEnd(ev) {
		var s = ev.tag.stats;
		s || (s = ev.tag.stats = {});
		s.closeEnd = getPosEnd(ev.event.parser);
		this.statsLastTagInit = null;
	}
};
function statsEvent(ev) {
	var fn = statsEventMap[ev.name];
	if (fn) { fn.call(this, ev); }
}

function HTMLTypeset() {
	this.errors = [];
	this.text = [];
	this.rootStyle = {
		fontFace: 'helvetica',
		fontSize: 10,
		fontBold: false,
		fontItalic: false,
		lineLeading: 15,
		paragraphLeading: 25,
		tolerance: void 0
	};
	this.scopeReset();
	this.treeEvent = this.treeEvent.bind(this);
}
HTMLTypeset.prototype = {
	errors: null,
	text: null,
	currentParagraph: null,
	currentScope: null,
	pathScope: null,
	currentSpan: null,
	rootStyle: null,
	reInline: /^em$|^span$|^strong$|^b$|^i$/i,
	reBold: /^strong$|^b$/i,
	reItalic: /^em$|^i$/i,
	ensureParagraph: function () {
		if (this.currentParagraph) { return; }
		var s = this.currentScope || this.rootStyle;
		this.currentParagraph = {
			leading: s.lineLeading,
			paragraphLeading: s.paragraphLeading,
			tolerance: s.tolerance,
			spans: []
		};
	},
	pushParagraph: function () {
		if (!this.currentParagraph) { return; }
		this.pushSpan();
		this.text.push(this.currentParagraph);
		this.currentParagraph = void 0;
	},
	ensureSpan: function () {
		this.ensureParagraph();
		if (this.currentSpan) { return; }
		var cs = this.currentScope;
		this.currentSpan = {
			fontFamily: [
				cs.fontFace,
				(cs.fontBold || cs.fontItalic)
					? (cs.fontBold ? 'bold' : '') + (cs.fontItalic ? 'italic' : '')
					: 'normal'
			].join('\t'),
			fontSize: cs.fontSize,
			text: ''
		};
	},
	normalizeWhitespace: function(str) {
		return str.replace(/[ \r\n\t]+/g,' ');
	},
	decodeString: function(str) {
		return str;
		// this function is meant to be replaced
		// with a function to decode html entities
	},
	processTextCommit: function(text) {
		return this.decodeString(this.normalizeWhitespace(text));
	},
	pushSpan: function () {
		if (!this.currentSpan) { return; }
		var sText = this.currentSpan.text;
		if (sText) {
			this.currentSpan.text = this.processTextCommit(sText);
			this.currentParagraph.spans.push(this.currentSpan);
		}
		this.currentSpan = void 0;
	},
	scopeParent: function () {
		var cs = this.currentScope;
		return cs && cs.parent || this.rootStyle;
	},
	scopePush: function (tag, style) {
		var scope = this.currentScope || this.rootStyle;
		this.currentScope = ObjectAssign({}, scope,
			style,
			{tagOpen: tag,
			tagClose: null,
			parent: this.currentScope});
		this.pathScope.push(this.currentScope);
	},
	scopePop: function (tag) {
		this.currentScope.tagClose = tag;
		this.currentScope = this.currentScope.parent;
		this.pathScope.pop();
	},
	scopeReset: function () {
		this.currentScope = void 0;
		this.pathScope = [];
	},
	treeTagOpen: function(tagName, tag) {
		var style = {};
		if (this.reBold.test(tagName)) { style.fontBold = true; }
		if (this.reItalic.test(tagName)) { style.fontItalic = true; }
		this.scopePush(tag || tagName, style);
		if (this.reInline.test(tagName)) {
			this.pushSpan();
		} else {
			this.pushParagraph();
		}
	},
	treeTagClose: function(tagName, tag) {
		this.scopePop(tag);
		if (this.reInline.test(tagName)) {
			this.pushSpan();
		} else {
			this.pushParagraph();
		}
	},
	processTextInput: function(text) {
		return text;
	},
	treeText: function(text) {
		this.ensureSpan();
		this.currentSpan.text += this.processTextInput(text);
	},
	treeOnError: function(ev) {
		this.errors.push(ev);
	},
	treeEvent: function(ev) {
		var evName = ev.name;
		var tag = ev.tag;
		var tagName = tag && tag.name;
		if (ev.error) { this.treeOnError(ev); }
		switch (evName) {
			case 'tagOpenEnd': this.treeTagOpen(tagName, tag); break;
			case 'tagCloseEnd': this.treeTagClose(tagName, tag); break;
			case 'text': this.treeText(ev.text); break;
		}
	},
	getResult: function() {
		this.pushParagraph();
		return {
			errors: this.errors,
			result: this.text
		};
	}
};

XMLParser.optDefault.tagStrictMap = htmlStrictTagMap;

TreeBuilder.optDefault.element = elementDefault;
TreeBuilder.optDefault.tagVoidMap = htmlVoidTagMap;

Printer.optDefault.tagVoidMap = htmlVoidTagMap;
Printer.optDefault.tagStrictMap = htmlStrictTagMap;

exports.HTMLTypeset = HTMLTypeset;
exports.Printer = Printer;
exports.TreeBuilder = TreeBuilder;
exports.TreeMatcher = TreeMatcher;
exports.XMLParser = XMLParser;
exports.elementDefault = elementDefault;
exports.elementDom = dom;
exports.elementSnabbdom = snabbdom;
exports.getParser = getParser;
exports.htmlStrictTagMap = htmlStrictTagMap;
exports.htmlVoidTagMap = htmlVoidTagMap;
exports.printerTransform = printertransform;
exports.testList = testList;
exports.treeRender = treerender;
exports.treeStats = statsEvent;

Object.defineProperty(exports, '__esModule', { value: true });

});
